/*
  This is a simple Xwindow that displays live video

  This one tries to use a lot of fast techniques
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include <string.h>
#include <visenv.h>
#include <visX.h>
#include <stdbool.h>

#define APPNAME "Cygnet Cropper\0"

/*
  Global Variables determining window and display
*/
Display *curDisplay;
Window   curWindow;
GC       curGC; /* graphics context */
XImage  *xim1, *xim2;
int     defdepth,bitmap_pad,defScreen;
int     bytesPerPixel;
Pixmap  gblPixmap;
Colormap curMap;
XColor theColor;
XSetWindowAttributes wAttr;
XGCValues gcValues;
char xbuffer[64] = APPNAME;



int Xsize=800;
int Ysize=600;

/*
  Global Variables for userland stuff
*/

unsigned char *ximdata;
unsigned char *xim2data;
bool boxed,go=true;
int bstart[2];
int bbox[4];

//prototypes
int initWindow(long nFrames);
void refreshWindow(void);
void closeWindow(long nFrames);
void adjustImage(Image *theImage);
void cropImage(Image *theImage);
/**********************************
 *       XLIB STUFF
 **********************************/

void refreshWindow(void){
  XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);  
  if(boxed){
    XDrawRectangle(curDisplay, curWindow, curGC, bbox[0], bbox[1],bbox[2] , bbox[3]);    
  }
  return;
}

void closeWindow(long nFrames)
{

   XCloseDisplay(curDisplay);

   return;
}

/****************************************
 *        MAIN LOOP
 ****************************************/

int main(int argc, char * argv[])
{
   bool done = false,newim=true;
   XEvent curEvent;
   char c;
   long event_mask = 0xFFFFFFFF;
   long nBuf = 2;
   long start;
   bool loop=true,clicking;
   char filename[MAXBUF],target[MAXBUF],buffer[MAXBUF],source[MAXBUF],list[MAXIMS][MAXBUF];
   int count=0,i=0,temp,required,rxreturn,ryreturn,wxreturn,wyreturn,oldbox[4],j=0;
   unsigned int mask;
   Window root, child;
   Image *theImage;
   XTextProperty xtp={xbuffer, 125, 8, strlen(xbuffer)};

   /* defaults */
   boxed=false;
   clicking=false;

   /* set required to the number of arguments that must be found..
      not sure what the best way to do this is. */
   required=3;
   while(loop){
     temp=getopt(argc,argv,"f:hs:t:");
     
     switch(temp){
     case 'f':
       strncpy(filename,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case 's':
       strncpy(source,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case 't':
       strncpy(target,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case -1:
       if(count>=required){
	 loop=false;
	 break;
       }
     case ':': case 'h': case '?':
       printf("This program crops and resizes images based on a user supplied bounding box.\n\n");
       printf("Usage: %s -f <image list> -s <image dir> -t <target dir>\n",argv[0]);
       printf("\t <image list> should contain a list of all cygnet photos, without their .ppm suffix\n");
       printf("\t (all photos should be in ppm format. use \"convert\" to switch between tif, jpg, and ppm)\n");
       exit(1);
       break;
       
     }
   }
   
   count=readfilenames(filename,list);
   /* initialize the window and video device */
   initWindow(nBuf);
   
   start = time(NULL);
   
   theImage=(Image *)malloc(sizeof(Image));
   theImage->image=NULL;
   theImage->color=true;

   printf("Done initializing\n");
   while(!done) {
     if(j>=count){
       //       done=true;
       //break;
     }
     
     // do any processing you need to do here
     // xim1->data points to the data (unsigned char *)
     if(go){
       if(newim){
	 printf("Loading image %s/%s.ppm...\n",source, list[j]);
	 boxed=false;
	 clicking=false;
	 sprintf(buffer,"%s/%s.ppm",source,list[j]);
	 if(theImage&&theImage->image)
	   free(theImage->image);
	 theImage->image=readPPM(&theImage->rows,&theImage->cols,&temp,buffer);
	 if(!theImage->image){
	   printf("Error opening file %s\n",buffer);
	   exit(-1);
	 }
	 theImage->imagesize=theImage->rows*theImage->cols;
	 theImage->color=true;
	 theImage->chroma=false;
	 if(theImage->rows>800||theImage->cols>600){
	   resizeIm(theImage,theImage->cols/2.0,theImage->rows/2.0);
	 }
	 newim=false;
	 //	 printf("Done loading.\n");
       }
       Xsize=theImage->cols;
       Ysize=theImage->rows;
       XDestroyImage(xim1);
       
       XDestroyWindow(curDisplay,curWindow);
       curWindow = XCreateWindow(curDisplay, DefaultRootWindow(curDisplay),
				 10, 10, Xsize, Ysize, 0, 
				 defdepth, InputOutput, 
				 DefaultVisual(curDisplay, defScreen),
				 0, &wAttr);
       
       /*
	* request mouse button and keypress events
	*/
       wAttr.event_mask = ButtonPressMask | KeyPressMask | ExposureMask | ResizeRedirectMask;
       XChangeWindowAttributes(curDisplay, curWindow, CWEventMask, &wAttr);
       
       /*
	* force it to appear on the screen
	*/
       XSetWMName(curDisplay, curWindow, &xtp);
       XStoreName(curDisplay,curWindow,APPNAME);
       
       /* create a new window */
       XMapWindow(curDisplay, curWindow);
       curGC = XCreateGC(curDisplay, curWindow, 0, &gcValues);
       XSetWindowColormap(curDisplay,curWindow,DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay)));
       curMap=DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay));
       bitmap_pad = defdepth == 24 ? 32 : defdepth;
       XParseColor(curDisplay,curMap,"#993333",&theColor);
       XAllocColor(curDisplay,curMap,&theColor);
       XSetForeground(curDisplay,curGC,theColor.pixel);
	 
       ximdata = malloc(Xsize * Ysize * bytesPerPixel);
       for(i=0;i<Xsize*Ysize*bytesPerPixel;i++)
	 ximdata[i] = 0x00;
       xim1 = XCreateImage(curDisplay, DefaultVisual(curDisplay, defScreen), defdepth, 
			   ZPixmap, 0, ximdata, Xsize, Ysize, bitmap_pad, 0);
       xim1->data=ConvertToX((Pixel *)theImage->image, xim1->data,Ysize, Xsize,bytesPerPixel);
       XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);
       if(boxed)
	 XDrawRectangle(curDisplay, curWindow, curGC, bbox[0], bbox[1],bbox[2] , bbox[3]);
       go=false;
     }  
     

     if(XCheckWindowEvent(curDisplay, curWindow, event_mask, &curEvent)) {
       
       switch (curEvent.type) {
	 
       case ButtonPress:
	 /*
	   access button by curEvent.xbutton.button (1,2,3)
	   access position by curEvent.xbutton.x and .y
	 */
	 if(!clicking){
	   bbox[0]=0;
	   bbox[1]=0;
	   bbox[2]=0;
	   bbox[3]=0;
	   clicking=true;
	   boxed=true;
	   bstart[0]=curEvent.xbutton.x;
	   bstart[1]=curEvent.xbutton.y;

	 }else{
	   /* stop drawing the bounding box */
	   if(bstart[0]<curEvent.xbutton.x){
	     bbox[0]=bstart[0];
	     bbox[2]=curEvent.xbutton.x;
	   }else{
	     bbox[0]=curEvent.xbutton.x;
	     bbox[2]=bstart[0];
	   }
	   if(bstart[1]<curEvent.xbutton.y){
	     bbox[1]=bstart[1];
	     bbox[3]=curEvent.xbutton.y;
	   }else{
	     bbox[1]=curEvent.xbutton.y;
	     bbox[3]=bstart[1];
	   }
	   clicking=false;
	   if(bbox[2]-bbox[0]<5||bbox[3]-bbox[1]<5)
	     boxed=false;
	 }
	 break;
	 
       case KeyPress:
	 /*
	   access string using XLookupString(*event,*char,numChars,NULL,NULL)
	 */
	 if(XLookupString((XKeyEvent *)&curEvent, &c, 1, NULL, NULL) == 1) {
	   
	   //	   printf("key code %02X\n", (int)c);
	   
	   switch(c) {

	   case 'p':
	     adjustImage(theImage);
	     //j++;
	     //go=true;
	     refreshWindow();
	     break;
	     
	   case 'c':
	     cropImage(theImage);
	     //j++;
	     go=true;
	     boxed=false;
	     refreshWindow();
	     break;
	   case 'n':
	     printf("Skipping\n");
	     go=true;
	     j++;
	     newim=true;
	     break;

	   case 'r':
	     printf("Redo\n");
	     go=true;
	     newim=true;
	     break;
	   case 'b':
	     printf("Back...\n");
	     go=true;
	     j--;
	     newim=true;
	     break;
	   case 's':
	     sprintf(buffer,"%s/%s.ppm",target, list[j]);
	     writePPM(theImage->image,theImage->rows,theImage->cols,255,buffer);
	     go=true;
	     j++;
	     newim=true;
	     break;
	     
	   case 'q':
	     done = True;
	     break;
	     
	   case 'i':
	     //move the crop-box up
	     if(bbox[1]>0&&boxed){
	       bbox[1]--;
	       refreshWindow();
	     }
	     //	     printf("bbox[0]: %d bbox[1]: %d bbox[2]: %d bbox[3] %d\n",bbox[0],bbox[1],bbox[2],bbox[3]);
	     break;
	   case 'm':
	     //move it down
	     if(bbox[3]-bbox[1]<theImage->rows-1&&boxed){
	       bbox[1]++;
	       refreshWindow();
	     }
	     //printf("bbox[0]: %d bbox[1]: %d bbox[2]: %d bbox[3] %d\n",bbox[0],bbox[1],bbox[2],bbox[3]);
	     break;
	   case 'k':
	     //take a step to the right
	     if(bbox[2]-bbox[0]<theImage->cols-1&&boxed){
	       bbox[0]++;
	       refreshWindow();
	     }
	     //printf("bbox[0]: %d bbox[1]: %d bbox[2]: %d bbox[3] %d\n",bbox[0],bbox[1],bbox[2],bbox[3]);
	     break;
	   case 'j':
	     //left
	     if(bbox[0]>0&&boxed){
	       bbox[0]--;
	       refreshWindow();
	     }
	     //printf("bbox[0]: %d bbox[1]: %d bbox[2]: %d bbox[3] %d\n",bbox[0],bbox[1],bbox[2],bbox[3]);
	     break;
	   default:
	     break;
	   }
	 }
	 break;
	 
       case Expose:
	 refreshWindow();
	 break;
	 
       case ResizeRequest:
	 XResizeWindow(curDisplay, curWindow, Xsize, Ysize);
	 //	 printf("NO RESIZE FOR YOU!\n");
	 break;
	 
       default:
	 break;
       }
     }

     /* process mouse movement for bounding box here */
     if(clicking){
       if(XQueryPointer(curDisplay, curWindow, &root, &child, &rxreturn, &ryreturn, 
			&wxreturn, &wyreturn, &mask)){
	 oldbox[0]=bbox[0];
	 oldbox[1]=bbox[1];
	 oldbox[2]=bbox[2];
	 oldbox[3]=bbox[3];
	 if(bstart[0]<wxreturn){
	   bbox[0]=bstart[0];
	   bbox[2]=wxreturn-bstart[0];
	 }else{
	   bbox[0]=wxreturn;
	   bbox[2]=bstart[0]-wxreturn;
	 }
	 if(bstart[1]<wyreturn){
	   bbox[1]=bstart[1];
	   bbox[3]=wyreturn-bstart[1];
	 }else{
	   bbox[1]=wyreturn;
	   bbox[3]=bstart[1]-wyreturn;
	 }
       }
       if(bbox[0]!=oldbox[0]||
	  bbox[1]!=oldbox[1]||
	  bbox[2]!=oldbox[2]||
	  bbox[3]!=oldbox[3])
	 refreshWindow();
     }

     //     usleep(100);
   }  
   
   
   closeWindow(nBuf);
   
   return(0);
}


void adjustImage(Image *theImage){
  int neww,newh,width,height;
  double curratio,targratio;

  neww=width=bbox[2]-bbox[0];
  newh=height=bbox[3]-bbox[1];
  
  targratio=105/127.0;
  curratio=width/(double)height;

  /*
  if(curratio>targratio){
    newh=width*(127.0/105);
  }else if(curratio<targratio){
    neww=height*(105.0/127.0);
  }
  */
  //  printf("Current ratio %f\n",curratio);
  curratio=(double)neww/(double)newh;
  
  while(curratio-targratio>0.00001||targratio-curratio>0.00001){  
    if(curratio-targratio<0){
      neww++;
      while(neww+bbox[0]>=theImage->cols){
	bbox[0]--;
      }
    
    }else{
      newh++;
      while(newh+bbox[1]>=theImage->rows){
	bbox[1]--;
      }
    
    }
    if(bbox[0]<0||bbox[1]<0){
      //  printf("D'oh. Starting points less than 0, no good. try again.\n");
      break;
    }
    curratio=(double)neww/(double)newh;
  }

  if(bbox[0]<0||bbox[1]<0){
    while(bbox[0]<0){
      bbox[0]++;
    }
    while(bbox[1]<0){
      bbox[1]++;
    }

    while(curratio-targratio>0.00001||targratio-curratio>0.00001){  
      if(curratio-targratio<0){
	newh--;
	//	while(neww+bbox[1]>=theImage->cols){
	// bbox[1]++;
	//}
	
      }else{
	neww--;
	//while(newh+bbox[0]>=theImage->rows){
	// bbox[0]++;
	//}
	
      }
      curratio=(double)neww/(double)newh;
    }
    
  }

  bbox[2]=neww;
  bbox[3]=newh;
  //printf("New ratio: %f target: %f\n", curratio,targratio);
}

void cropImage(Image *theImage){

  //printf("%d %d %d %d\n",bbox[0],bbox[1],bbox[2],bbox[3]);
  if(bbox[2]>theImage->cols||bbox[3]>theImage->rows||bbox[1]>theImage->cols||bbox[3]>theImage->rows){
    printf("D'oh. You need to adjust the program some more.\n");
    exit(1);
  }
  //  printf("Current Size: %d %d\n",theImage->cols,theImage->rows);  
  //printf("Cropping to %d %d\n",bbox[2],bbox[3]);
  cropArbIm(theImage,bbox[0],bbox[2]+bbox[0],bbox[1],bbox[3]+bbox[1]);
  // printf("Resizing...\n");
  // printf("Current Size: %d %d\n",theImage->cols,theImage->rows);
  resizeIm(theImage,105,127);
  //printf("bbox[0]: %d bbox[1]: %d bbox[2]: %d bbox[3] %d\n",bbox[0],bbox[1],bbox[2],bbox[3]);
  //  writePPM(theImage->image,theImage->rows,theImage->cols,255,"test.ppm");
  //  go=true;
}


int initWindow(long nFrames) {
   int i;

   XTextProperty xtp={xbuffer, 125, 8, strlen(xbuffer)};
   /*
    * connect to the X server.  uses the server specified in the
    * DISPLAY environment variable
    */
   curDisplay = XOpenDisplay((char *) NULL);
   if ((Display *) NULL == curDisplay) {
      fprintf(stderr, "Could not open display.\n");
      exit(-1);
   }
   
   /*
    * begin to create a window
    */
   printf("creating window\n");
   defdepth = DefaultDepth(curDisplay,0);
   bytesPerPixel = defdepth/8;
   bytesPerPixel = bytesPerPixel == 3 ? 4 : bytesPerPixel;
   printf("default depth %d\n", defdepth);

   defScreen = DefaultScreen(curDisplay);
   printf("Default screen %d\n", defScreen);

   curWindow = XCreateWindow(curDisplay, DefaultRootWindow(curDisplay),
			     10, 10, Xsize, Ysize, 0, 
			     defdepth, InputOutput, 
			     DefaultVisual(curDisplay, defScreen),
			     0, &wAttr);

   printf("Using depth %d\n", DefaultDepth(curDisplay, 0));

   /*
    * request mouse button and keypress events
    */
   wAttr.event_mask = ButtonPressMask | KeyPressMask | ExposureMask | ResizeRedirectMask;
   XChangeWindowAttributes(curDisplay, curWindow, CWEventMask, &wAttr);
  
   /*
    * force it to appear on the screen
    */
   XSetWMName(curDisplay, curWindow, &xtp);
   XStoreName(curDisplay,curWindow,APPNAME);


   /*
    * create a graphics context.  this stores a drawing state; stuff like
    * current color and line width.  this gc is for drawing into our
    * window. 
    */
   printf("creating graphics context\n");
   curGC = XCreateGC(curDisplay, curWindow, 0, &gcValues);

   XSetWindowColormap(curDisplay,curWindow,DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay)));
   curMap=DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay));
   //   curScreen=DefaultScreen(curDisplay);
   /*
    * create an XImage for holding image data
    */

   bitmap_pad = defdepth == 24 ? 32 : defdepth;

   ximdata = malloc(Xsize * Ysize * bytesPerPixel);
   for(i=0;i<Xsize*Ysize*bytesPerPixel;i++)
     ximdata[i] = 0x00;
   xim1 = XCreateImage(curDisplay, DefaultVisual(curDisplay, defScreen), defdepth, 
		       ZPixmap, 0, ximdata, Xsize, Ysize, bitmap_pad, 0);

   //   printf("creating frame 1 with depth %d\n", defdepth);
   
   printf("exiting init window\n");
   return(bytesPerPixel);
}

