/*
  This is a simple Xwindow that displays live video

  This one tries to use a lot of fast techniques
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <ppmIO.h>
#include <string.h>
#include <visenv.h>
#include <keyboard.h>
#include <visX.h>

/*
  Global Variables determining window and display
*/
Display *curDisplay;
Window   curWindow;
GC       curGC; /* graphics context */
XImage  *xim1, *xim2;
int     defdepth;
int     bytesPerPixel;
Pixmap  gblPixmap;

#define Xsize 320
#define Ysize 240

/*
  Global Variables for userland stuff
*/
unsigned char *ximdata;

int threshpad;
int grass;
Pixel thresh;
bool setthresh;

int initWindow(long nFrames);
void refreshWindow(void);
void closeWindow(long nFrames);
Image *processImage(Image *theImage);

int main(void)
{
   Bool   done = False;
   XEvent curEvent;
   char c;
   long event_mask = 0xFFFFFFFF;
   long nBuf = 2;
   long nFrames = 0;
   long start, end;
   Pixel *ppm;
   char ppmfilename[128];
   int ppmcounter = 0;
   Point_t threshPoint;
   Image *img,*tmpimg,*t;
   setthresh=false;
   grass=5;
   threshpad=70;
   thresh=(Pixel){96,84,74};
   /* initialize the window and video device */
   initWindow(nBuf);

   start = time(NULL);

   CaptureVideo(0);
   CaptureVideo(1);

   //   img=newImage(Ysize,Xsize,false,true);

   img=(Image *)malloc(sizeof(Image));
   img->cols=Xsize;
   img->rows=Ysize;
   img->color=true;
   img->chroma=false;
   img->imagesize=img->rows*img->cols;

   //   img->image=(Pixel *)malloc(sizeof(Pixel)*img->imagesize);

   // free(img->image);

   while(!done) {
     // get another picture
     SyncVideo(0);

     // do any processing you need to do here
     // xim1->data points to the data (unsigned char *)

     img->image=ConvertToPPM(xim1->data,Xsize,Ysize,bytesPerPixel);
     img=processImage(img);
     //xim1->data=
     ConvertToX((Pixel *)img->image, xim1->data,Xsize, Ysize,bytesPerPixel);
     free(img->image);
     XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);
     CaptureVideo(0);
     nFrames++;
    
     SyncVideo(1);

     // do any processing you need to do here
     // xim2->data points to the data (unsigned char *)

     img->image=ConvertToPPM(xim2->data,Xsize,Ysize,bytesPerPixel);
     img=processImage(img);
     //xim2->data=
     ConvertToX((Pixel *)img->image, xim2->data, Xsize, Ysize,bytesPerPixel);
     free(img->image);
     XPutImage(curDisplay, curWindow, curGC, xim2, 0, 0, 0, 0, Xsize, Ysize);
     CaptureVideo(1);
     nFrames++;


     if(XCheckWindowEvent(curDisplay, curWindow, event_mask, &curEvent)) {

       switch (curEvent.type) {

       case ButtonPress:
	 /*
	   access button by curEvent.xbutton.button (1,2,3)
	   access position by curEvent.xbutton.x and .y
	 */
	 if(setthresh){
	   if(curEvent.xbutton.button == 1){
	     tmpimg=(Image *)malloc(sizeof(Image));
	     tmpimg->rows=Ysize;
	     tmpimg->cols=Xsize;
	     tmpimg->imagesize=Xsize*Ysize;
	     tmpimg->color=true;
	     tmpimg->image = ConvertToPPM(xim1->data, Xsize, Ysize, bytesPerPixel);
	     gaussianIm(tmpimg);
	     cropXYIm(tmpimg, 16, 12);
	     
	     t=theImage;
	     tmpimg=chromaticityIm(t);
	     free(t->image);
	     free(t);
	     
	     threshByPaddedColor(tmpimg,thresh, threshpad);
	     if(grass>0){
	       grassFire(tmpimg, (Pixel){0,0,0}, grass, false, false);
	       grassFire(tmpimg, (Pixel){0,0,0}, grass, true, true);
	     }
	     
	     resizeIm(tmpimg,tmpimg->cols/2,tmpimg->rows/2);
	     resizeIm(tmpimg,tmpimg->cols/2,tmpimg->rows/2);
	     resizeIm(tmpimg,tmpimg->cols/2,tmpimg->rows/2);
	     
	     /*resize to create the display image*/
	     resizeIm(tmpimg,tmpimg->cols*2,tmpimg->rows*2);
	     resizeIm(tmpimg,tmpimg->cols*2,tmpimg->rows*2);
	     /*resizeIm(tmpimg,tmpimg->cols*2,tmpimg->rows*2);*/
	     resizeIm(tmpimg,Xsize,Ysize);
	     thresh=tmpimg->image[curEvent.xbutton.y*tmpimg->cols+curEvent.xbutton.x];
	     free(tmpimg->image);
	     free(tmpimg);
	     printf("Threshold changed to %d %d %d\n",(int)thresh.r,(int)thresh.g,(int)thresh.b);
	   }else{
	     printf("Threshold unchanged.\n");
	   }
	   setthresh=false;
	   break;
	 }
	 if(curEvent.xbutton.button == 1) {
	   // write out the current frame to a ppm
	   SyncVideo(0);

	   ppm = ConvertToPPM(xim1->data, Xsize, Ysize, bytesPerPixel);
	   sprintf(ppmfilename, "frame.%03d.ppm", ppmcounter++);
	   writePPM(ppm, Ysize, Xsize, 255, ppmfilename);
	   
	   free(ppm);
	   
	   CaptureVideo(0);
	   SyncVideo(1);
	   CaptureVideo(1);
	   
	 }
	 else if(curEvent.xbutton.button == 2) {
	   threshPoint[0]=curEvent.xbutton.x;
	   threshPoint[1]=curEvent.xbutton.y;
	 }
	 break;
	 
       case KeyPress:
	 /*
	   access string using XLookupString(*event,*char,numChars,NULL,NULL)
	 */
	 if(XLookupString((XKeyEvent *)&curEvent, &c, 1, NULL, NULL) == 1) {
	   switch(c) {
	   case 'q':
	     done = True;
	     break;

	   case 'g':
	     grass++;
	     printf("Grassfire level: %d\n",grass);
	     break;	     
	   case 'f':
	     grass--;
	     printf("Grassfire level: %d\n",grass);
	     break;
	   case '+':
	     threshpad++;
	     printf("Threshpad level: %d\n",threshpad);
	     break;
	   case '-':
	     threshpad--;
	     printf("Threshpad level: %d\n",threshpad);
	     break;
	   case 't':
	     printf("Left click to set threshold(other buttons will cancel\n");
	     setthresh=true;
	     break;
	   case '0':
	   case '1':
	   case '2':
	   case '3':
	     SyncVideo(0);
	     SyncVideo(1);

	     SetVideoChannel(c - '0');

	     CaptureVideo(0);
	     CaptureVideo(1);
	     break;
	   default:
	     break;
	   }
	 }
	 break;

       case Expose:
	 refreshWindow();

       default:
	 break;
       }
     }
   }
   end = time(NULL);

   printf("Frame rate = %.2f/s\n", (double)nFrames / (double)(end - start));

   closeWindow(nBuf);

   return(0);
}

/*********
 * image processing
 **********/

/*
 * the network: 36x27 input nodes.
 * 15 hidden nodes, layer 1
 * 8 hidden nodes, layer 2
 * 4 output nodes
 */
Image * processImage(Image *theImage){
  Image *tempimg;

  gaussianIm(theImage);
  cropXYIm(theImage, 16, 12);
  
  tempimg=theImage;
  theImage=chromaticityIm(tempimg);
  free(tempimg->image);
  free(tempimg);

  if(!setthresh){
    threshByPaddedColor(theImage,thresh, threshpad);
    if(grass>0){
      grassFire(theImage, (Pixel){0,0,0}, grass, false, false);
      grassFire(theImage, (Pixel){0,0,0}, grass, true, true);
    }
  }
  resizeIm(theImage,theImage->cols/2,theImage->rows/2);
  resizeIm(theImage,theImage->cols/2,theImage->rows/2);
  resizeIm(theImage,theImage->cols/2,theImage->rows/2);
  
  /*resize to create the display image*/
  resizeIm(theImage,theImage->cols*2,theImage->rows*2);
  resizeIm(theImage,theImage->cols*2,theImage->rows*2);
  /*resizeIm(theImage,theImage->cols*2,theImage->rows*2);*/
  resizeIm(theImage,Xsize,Ysize);
  
  return theImage;
}


/**************
 * X functions
 **************/

int initWindow(long nFrames) {
   int defScreen;
   XSetWindowAttributes wAttr;
   XGCValues gcValues;
   int bitmap_pad;
   long height = Ysize, width = Xsize;
   long *offsets;
   char buffer[64] = "Live Video";
   XTextProperty xtp = {buffer, 125, 8, strlen(buffer)};

   offsets = (long *)malloc(sizeof(long) * nFrames);
   
   /*
    * connect to the X server.  uses the server specified in the
    * DISPLAY environment variable
    */
   curDisplay = XOpenDisplay((char *) NULL);
   if ((Display *) NULL == curDisplay) {
      fprintf(stderr, "xskeleton:  could not open display.\n");
      exit(-1);
   }
   
   /*
    * begin to create a window
    */
   printf("creating window\n");
   defdepth = DefaultDepth(curDisplay,0);
   bytesPerPixel = defdepth/8;
   bytesPerPixel = bytesPerPixel == 3 ? 4 : bytesPerPixel;
   printf("default depth %d\n", defdepth);

   defScreen = DefaultScreen(curDisplay);
   printf("Default screen %d\n", defScreen);

   curWindow = XCreateWindow(curDisplay, DefaultRootWindow(curDisplay),
			     10, 10, Xsize, Ysize, 0, 
			     defdepth, InputOutput, 
			     DefaultVisual(curDisplay, defScreen),
			     0, &wAttr);

   printf("Using depth %d\n", DefaultDepth(curDisplay, 0));

   /*
    * request mouse button and keypress events
    */
   wAttr.event_mask = ButtonPressMask | KeyPressMask | ExposureMask;
   XChangeWindowAttributes(curDisplay, curWindow, CWEventMask, &wAttr);

   /*
    * force it to appear on the screen
    */
   XSetWMName(curDisplay, curWindow, &xtp);
   XMapWindow(curDisplay, curWindow);

   /*
    * create a graphics context.  this stores a drawing state; stuff like
    * current color and line width.  this gc is for drawing into our
    * window. 
    */
   printf("creating graphics context\n");
   curGC = XCreateGC(curDisplay, curWindow, 0, &gcValues);

   XSetWindowColormap(curDisplay,curWindow,DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay)));

   /*
    * create an XImage for holding image data
    */

   printf("Allocating space for %ld frames with %d bytes per pixel\n", nFrames, bytesPerPixel);
   ximdata = OpenVideoDevice(&height, &width, bytesPerPixel, nFrames, 1, offsets);

   bitmap_pad = defdepth == 24 ? 32 : defdepth;

   printf("creating frame 1 with depth %d\n", defdepth);
   xim1 = XCreateImage(curDisplay, DefaultVisual(curDisplay, defScreen), defdepth, 
		       ZPixmap, 0, &(ximdata[offsets[0]]), Xsize, Ysize, bitmap_pad, 0);

   if(nFrames > 1) {
     printf("creating frame 2 with depth %d\n", defdepth);
     xim2 = XCreateImage(curDisplay, DefaultVisual(curDisplay, defScreen), defdepth, 
			 ZPixmap, 0, &(ximdata[offsets[1]]), Xsize, Ysize, 
			 bitmap_pad, 0);
   }

   free(offsets);

   printf("exiting init window\n");
   return(bytesPerPixel);
}

void refreshWindow()
{
  // refresh the window here
  XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);

  return;
}

void closeWindow(long nFrames)
{
   CloseVideoDevice(ximdata, Ysize, Xsize, bytesPerPixel, nFrames);
   
   XCloseDisplay(curDisplay);

   return;
}

