#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ppmIO.h>
#include <visX.h>

void InsertPixel(int r, int g, int b, long y, long x, unsigned char *im, const long cols, const long rows, const long bpp);
void InsertPixel(int r, int g, int b, long y, long x, unsigned char *im, const long cols, const long rows, const long bpp) {
  unsigned char *utmp;
  short *stmp;

  if(y < 0 || y >= rows)
    return;
  if(x < 0 || x >= cols)
    return;

  switch(bpp) {
  case 1:
    im[y*cols + x] = (r + g + g + b) >> 2;
    break;
  case 2:
    stmp = (short *)&im[(y*cols + x)*bpp];
    *stmp = ((r & 0xF8) << 8) + ((g & 0xFC) << 3) + (b >> 3);
    break;
  case 3:
    utmp = &im[(y*cols + x)*bpp];
    *(utmp) = (unsigned char)b;
    *(utmp+1) = (unsigned char)g;
    *(utmp+2) = (unsigned char)r;
    break;
  case 4:
    utmp = &im[(y*cols + x)*bpp];
    *(utmp) = (unsigned char)b;
    *(utmp+1) = (unsigned char)g;
    *(utmp+2) = (unsigned char)r;
    break;
  default:
    break;
  }
}

void InsertBox(int r, int g, int b, long box[], unsigned char *data, const long cols, const long rows, const long bpp);
void InsertBox(int r, int g, int b, long box[], unsigned char *data, const long cols, const long rows, const long bpp) {
  long i;

  // put in the top and bottom lines
  for(i=box[1];i<=box[3];i++) {
    InsertPixel(r, g, b, box[0], i, data, cols, rows, bpp);
    InsertPixel(r, g, b, box[2], i, data, cols, rows, bpp);
  }

  // put in the left and right lines
  for(i=box[0];i<box[2];i++) {
    InsertPixel(r, g, b, i, box[1], data, cols, rows, bpp);
    InsertPixel(r, g, b, i, box[3], data, cols, rows, bpp);
  }
}

Pixel *ConvertToPPM(unsigned char *data, const long cols, const long rows, const int depth);
Pixel *ConvertToPPM(unsigned char *data, const long cols, const long rows, const int depth) {
  long i;
  Pixel *ppm;
  unsigned char *tmp;
  short *stmp;

  ppm = (Pixel *)malloc(sizeof(Pixel) * rows * cols);
  if(!ppm)
    return(NULL);

  // convert it to a ppm
  switch(depth) {
  case 1:
    for(i=0;i<rows * cols;i++) {
      ppm[i].r = ppm[i].g = ppm[i].b = data[i];
    }
    break;
  case 2:
    // need to convert from 16 bit 5/6/5 to ppm format
    stmp = (short *)data;
    for(i=0;i<rows * cols;i++,stmp++) {
      ppm[i].r = ((*stmp) & 0xF800) >> 8;
      ppm[i].g = ((*stmp) & 0x07C0) >> 3;
      ppm[i].b = ((*stmp) & 0x001F) << 3;
    }
    break;
  case 3:
    memcpy((void *)ppm, (const void *)data, (size_t)rows * cols * 3 );
    break;
  case 4:
    tmp = (unsigned char *)data;
    for(i=0;i<rows * cols;i++,tmp += 4) {
      ppm[i].r = tmp[2];
      ppm[i].g = tmp[1];
      ppm[i].b = tmp[0];
    }
    break;
  default:
    ppm = NULL;
    break;
  }

  return(ppm);
}

unsigned char *ConvertToX(Pixel *data, unsigned char *output,const long cols, const long rows, const int depth);
unsigned char *ConvertToX(Pixel *data, unsigned char *output,const long cols, const long rows, const int depth) {
  long i;

  unsigned char *tmp;

  if(depth!=4){
    printf("Its not 4...Fix the ConvertToX function\n");
    exit(-1);
  }

  tmp=output;
  //  for(i=0;i<rows * cols;i++,tmp+=4) {
  /* for(i=0;i<rows * cols;i++) {

   tmp[i*4+2]=data[i].r;

   tmp[i*4+1]=data[i].g;

   tmp[i*4+0]=data[i].b;
   
   }*/

  for(i=0;i<rows * cols;i++,tmp += 4) {
    tmp[2]=data[i].r;
    tmp[1]=data[i].g;
    tmp[0]=data[i].b;
  }
  
  return output;
}

Pixel *ConvertToPPM2(unsigned char *data, Pixel *buffer, const long cols, const long rows, const int depth) {
  long i;
  Pixel *ppm;
  unsigned char *tmp;
  short *stmp;

  ppm=buffer;

  // convert it to a ppm
  switch(depth) {
  case 1:
    for(i=0;i<rows * cols;i++) {
      ppm[i].r = ppm[i].g = ppm[i].b = data[i];
    }
    break;
  case 2:
    // need to convert from 16 bit 5/6/5 to ppm format
    stmp = (short *)data;
    for(i=0;i<rows * cols;i++,stmp++) {
      ppm[i].r = ((*stmp) & 0xF800) >> 8;
      ppm[i].g = ((*stmp) & 0x07C0) >> 3;
      ppm[i].b = ((*stmp) & 0x001F) << 3;
    }
    break;
  case 3:
    memcpy((void *)ppm, (const void *)data, (size_t)rows * cols * 3 );
    break;
  case 4:
    tmp = (unsigned char *)data;
    for(i=0;i<rows * cols;i++,tmp += 4) {
      ppm[i].r = tmp[2];
      ppm[i].g = tmp[1];
      ppm[i].b = tmp[0];
    }
    break;
  default:
    ppm = NULL;
    break;
  }

  return(ppm);
}


Pixel *ConvertToRPM(unsigned char *data, const long cols, const long rows, const int depth);
Pixel *ConvertToRPM(unsigned char *data, const long cols, const long rows, const int depth) {
  long i;
  Pixel *ppm;
  unsigned char *tmp;
  short *stmp;

  ppm = (Pixel *)malloc(sizeof(Pixel) * rows * cols);
  if(!ppm)
    return(NULL);

  // convert it to a ppm
  switch(depth) {
  case 1:
    for(i=0;i<rows * cols;i++) {
      ppm[i].r = ppm[i].g = ppm[i].b = data[i];
    }
    break;
  case 2:
    // need to convert from 16 bit 5/6/5 to ppm format
    stmp = (short *)data;
    for(i=0;i<rows * cols;i++,stmp++) {
      ppm[i].b = ((*stmp) & 0xF800) >> 8;
      ppm[i].g = ((*stmp) & 0x07C0) >> 3;
      ppm[i].r = ((*stmp) & 0x001F) << 3;
    }
    break;
  case 3:
    memcpy((void *)ppm, (const void *)data, (size_t)rows * cols * 3 );
    break;
  case 4:
    tmp = (unsigned char *)data;
    for(i=0;i<rows * cols;i++,tmp += 4) {
      ppm[i].b = tmp[2];
      ppm[i].g = tmp[1];
      ppm[i].r = tmp[0];
    }
    break;
  default:
    ppm = NULL;
    break;
  }

  return(ppm);
}


