#include <visenv.h>

/****************
 * Operators:
 * Each function in this file
 * should return some sort of
 * generic type, ie, void *
 * These really generate feature
 * information,
 ****************/

Feature * freeFeature(Feature *input){

  switch(input->type){

  case DCHAIN_1:
  case DCHAIN_2:
    free(input->data);
    break;

  case CHAIN_CODE:
    freeChainCode((ChainCode *)input->data);
    free(input);
    break;

  case PERCENT:
    free(input->data);
    free(input);
    break;
    
  default:
    free(input->data);
    free(input);
    break;
  }
 
  return NULL;
}

/**************
 * chaincode functions
 *************/

/*buildChainCode returns a linked list of ChainLinks, given an
  image, a starting location on or near the border of a region,
  and a Pixel value for the background color */
Feature * buildChainCode(Image *image, Box_t bbox, Pixel bgVal){

  int i,j,loc,firstx=-1,firsty=-1,locx,locy,direction=6,nextx=-1,nexty=-1,xstart,ystart,height,width,counter=0;
  ChainCode * theCode;
  ChainLink * current;
  Feature *output;
  Image *labeled;

  xstart=bbox[1];
  ystart=bbox[0];
  height=bbox[2]-bbox[0];
  width=bbox[3]-bbox[1];

  /* malloc and initialize a temporary image for labels */
  labeled=newImage(image->rows,image->cols,false,false);
  
  // find the border, if we're not on it
  // from the start pixel to the end of the bounding box
  // or the end of the image,
  for(i=ystart;(i<ystart+height)&&(i<image->rows);i++){
    for(j=xstart;(j<xstart+width)&&(j<image->cols);j++){    

      loc=(i*image->cols)+j;
      
      if(!checkPixel(image,loc,bgVal)){
	firsty=i;
	firstx=j;
	labeled->gimage[firsty*image->cols+firstx]=(unsigned char)255;
	break;
      }
      direction=0;
    }
    if(firstx>-1){
      break;
    }
    direction=6;
  }
  
  //if start hasn't been reset, there's nothing here.
  if(firstx==-1){
    fprintf(stderr, "No foreground region found in bounding box! in buildChainCode()\n");
    return NULL;
  }
  
  theCode = (ChainCode *) malloc(sizeof(ChainCode));
  if(!theCode){
    fprintf(stderr,"Fatal error: out of memory in buildChainCode(): ChainCode allocation\n");
    exit(-1);
  }
  theCode->length=1;
  theCode->x=xstart;
  theCode->y=ystart;
  theCode->height=height;
  theCode->width=width;

  locx=firstx;
  locy=firsty;
  
  current=newLink(firstx,firsty);

  theCode->first=current;
 
  do{
   
    if(counter<4)
      counter++;

    direction-=3;

    for(i=0;i<8;i++,direction++){
      if(direction<0)
	direction+=8;
      else if(direction>=8)
	direction-=8;

      switch(direction){
      case 0: case 1: case 7:
	nextx=locx+1;
	break;
      case 2: case 6:
	nextx=locx;
	break;
      case 3: case 4: case 5:
	nextx=locx-1;
	break;
      default:
	fprintf(stderr,"Bogus direction in buildchaincode()\n");
	break;
      }
      switch(direction){
      case 0: case 4:
	nexty=locy;
	break;
      case 1: case 2: case 3:
	nexty=locy-1;
	break;
      case 5: case 6: case 7:
	nexty=locy+1;
	break;
      default:
	fprintf(stderr,"Bogus direction %d in buildChainCode()\n",direction);
	break;
      }
      
      // create a new link
      if((!checkPixel(image,nexty*image->cols+nextx,bgVal))&&
	 (labeled->gimage[nexty*image->cols+nextx]!=(unsigned char)255)&&
	 (is8Boundary(image,nextx,nexty,bgVal))){

	current->direction=direction;
	current->next=newLink(nextx,nexty);
	current->next->previous=(struct ChainLink_t *)current;
	current=(ChainLink *)current->next;
	locx=nextx;
	locy=nexty;
	theCode->length++;
	break;

	//we've returned to the start.
      }else if(nextx==firstx&&nexty==firsty){
	theCode->first->previous=(struct ChainLink_t *)current;
	theCode->last=current;
	current->next=theCode->first;
	current->direction=direction;
	locx=nextx;
	locy=nexty;
	break;
      }
    }

  }while(locx!=firstx||locy!=firsty);

  if(!theCode->last||!theCode->first->next){
    freeChainCode(theCode);
    return NULL;
  }

  // printf("locx %d locy %d firstx %d firsty %d\n",locx,locy,firstx,firsty);

  if(counter<4){
    freeChainCode(theCode);
    return NULL;
  }
  free(labeled->gimage);
  free(labeled);
  
  output=(Feature *)malloc(sizeof(Feature));
  if(!output){
    fprintf(stderr,"Fatal error: Out of memory in percentFilled()\n");
    exit(-1);
  }
  output->type=CHAIN_CODE;
  output->data=theCode;


  return output;
}


ChainLink *newLink(int locx, int locy){
  
  ChainLink *temp;

  temp=(struct ChainLink_t *)malloc(sizeof(ChainLink));
  
  if(!temp){
    fprintf(stderr,"Fatal error: out of memory in buildChainCode(): Link allocation, loc %d %d\n",locx,locy);
    exit(-1);
  }
  
  temp->x=locx;
  temp->y=locy;
  temp->next=NULL;
  temp->previous=NULL;
  return temp;
}

Feature *derivativeCode(ChainCode *code,FeatureType type){

  ChainCode *dcode;
  ChainLink *current,*temp;
  Feature *output;

  if(type!=DCHAIN_1&&type!=DCHAIN_2){
    fprintf(stderr, "Fix the bugs in your damn code. derivativeCode type is wrong.\n");
    exit(-1);
  }

  if(!code){
    fprintf(stderr,"Passed null pointed instead of ChainCode in derivativeCode()\n");
    return NULL;
  }

  dcode=(ChainCode *)malloc(sizeof(ChainCode));
  if(!dcode){
    fprintf(stderr,"Fatal error: out of memory in derivativeCode\n");
    exit(-1);
  }

  current=code->first;

  temp=NULL;
  temp=(ChainLink *)malloc(sizeof(ChainLink));
  if(!temp){
    fprintf(stderr,"Fatal error: out of memory in derivativeCode\n");
    exit(-1);
  }
  
  dcode->first=temp;
  temp->direction=current->direction-current->next->direction;
  if(temp->direction<0){
    temp->direction+=8;
  }else if(temp->direction>7){
    temp->direction-=8;
  }
  current=current->next;
  do{
    temp->next=newLink(0,0);

    temp->next->previous=temp;
    temp=temp->next;

    temp->direction=current->direction-current->next->direction;
    if(temp->direction<0){
      temp->direction+=8;
    }else if(temp->direction>7){
      temp->direction-=8;
    }  
    current=current->next;    
  }while(current!=code->first);

  dcode->last=temp->previous;

  output=(Feature *)malloc(sizeof(Feature));
  if(!output){
    fprintf(stderr,"Fatal error: Out of memory in percentFilled()\n");
    exit(-1);
  }
  output->type=type;
  output->data=dcode;

  return output;
}

void * freeChainCode(ChainCode *code){
  ChainLink *cur;

  int i=0,temp=code->length;

  if(!code){
    fprintf(stderr,"Passed null pointer in freeChainCode. Oops\n");
    return NULL;
  }
  
  if(!code->first){
     //||!code->first->next){
    fprintf(stderr,"Corrupt Chain Code\n");

    free(code);
    return NULL;
  }
  if(code->first->previous){
    code->first->previous->next=NULL;
  }
  cur=code->first;
  while(cur->next){

    cur=cur->next;
    free(cur->previous);
    i++;
    cur->previous=NULL;
  }

  free(cur);
  i++;
  free(code);

  if(i<temp){
    printf("LEAK!!!\n");
  }

  return NULL;

}

void drawChain(ChainCode * theCode,Image *image, Pixel color){

  ChainLink * cur;
  if(!theCode){
    printf("Passed a null pointer instead of a chain code in drawChain!\n");
    return;
  }
  cur=theCode->first;
  if(!cur||!cur->next){
    printf("Empty ChainCode, Bitch!\n");
    return;
  }
  do{
    setPixel(image,cur->x+cur->y*image->cols,color);
    if(cur->next){
      cur=(ChainLink*)cur->next;
    }else{
      printf("Bad chaincode-- no loop.\n");
      break;
    }
  }while(cur!=theCode->first);

}


void chainCodeToArray(Feature input){

  float *temp;
  int i=0;
  ChainLink *current;
  ChainCode *code;

  if(input.type!=CHAIN_CODE){
    fprintf(stderr,"Given something not a chain code in chainCodeToArray()\n");
    return;
  }

  if(input.data){
    code=(ChainCode *)input.data;
    if(code->first){
      if(!code->last){
	fprintf(stderr,"Bad chain code in chainCodeToArray()\n");
	return;
      }
    }else{
      fprintf(stderr,"Bad chain code in chainCodeToArray()\n");
      return;
    }
  }else{
    fprintf(stderr,"Bad chain code in chainCodeToArray()\n");
    return;
  }

  
  temp=calloc(8,sizeof(float));

  current=(ChainLink *)code->first;

  while(current&&i<code->length){
    temp[current->direction]+=1;
    current=current->next;
    i++;
  }
  
  for(i=0;i<code->length;i++){
    temp[i]=temp[i]/(float)code->length;
  }

  input.type=CHAIN_CODE_AR;

  input.data=temp;
  
  freeChainCode(code);
  
}

/************************
 * % bounding box filled
 ************************/

Feature * percentFilled(Image *image, Box_t bbox, Pixel bgVal){

  int i,j,k,sum,x,y,width,height;
  Feature *output;

  x=bbox[1];
  y=bbox[0];
  height=bbox[2]-bbox[0];
  width=bbox[3]-bbox[1];
  
  for(k=0,sum=0,i=y;(i<y+height)&&(i<image->rows);i++){
    for(j=x;(j<x+width)&&(j<image->cols);j++,k++){    
      if(!checkPixel(image,i*image->cols+j,bgVal)){
	sum++;
      }
    }
  }

  output=(Feature *)malloc(sizeof(Feature));
  if(!output){
    fprintf(stderr,"Fatal error: Out of memory in percentFilled()\n");
    exit(-1);
  }
  output->type=PERCENT;
  output->data=(double *)malloc(sizeof(double));
  if(!output->data){
    fprintf(stderr,"Fatal error: Out of memory in percentFilled()\n");
    exit(-1);
  }

  *(double*)output->data=(double)sum/(double)k;
 
  return output;

}

/*********
 * bounding box
 *********/

void drawBoundingBox(Image *image, Box_t bbox, Pixel color){
  
  int xstart,ystart,width,height,i,j;
  xstart=bbox[1];
  ystart=bbox[0];
  height=bbox[2]-bbox[0];
  width=bbox[3]-bbox[1];
  
  for(i=ystart;i<height+ystart;i++){
    for(j=xstart;j<width+xstart;j++){

      setPixel(image,j+i*image->cols,color);
    }
  }
}


/******************
 * Grassfire
 ******************/

/* invert flag will operate on background pixels rather than foreground */
void grassFire(Image * image, Pixel bgVal, short thresh, bool eight, bool invert){

  short *tmp,val; 
  int i,j;

  tmp=(short *)calloc(image->imagesize,sizeof(short));

  if(!tmp){
    fprintf(stderr,"Fatal error: out of memory in grassFire8().\n");
    exit(-1);
  }

  /* pass 1 */
  for(i=0;i<image->rows;i++){
    for(j=0;j<image->cols;j++){
      
      //XNOR
      if((i*image->cols+j)>image->imagesize){
	printf("Overflow!\n");
	exit(-1);
      }

      if(checkPixel(image,i*image->cols+j,bgVal)!=invert){
	tmp[i*image->cols+j]=0;
      }else{
	if(eight){
	  tmp[i*image->cols+j]=upLeftMin8(tmp,i,j,image->rows,image->cols)+1;  
	}else{
	  tmp[i*image->cols+j]=upLeftMin4(tmp,i,j,image->rows,image->cols)+1;
	}
      }
    }
  }

  /* pass 2 */

  for(i=image->rows-1;i>=0;i--){
    for(j=image->cols-1;j>=0;j--){

      if(checkPixel(image,i*image->cols+j,bgVal)!=invert){
	tmp[i*image->cols+j]=0;
      }else{
	if(eight){
	  val=downRightMin8(tmp,i,j,image->rows,image->cols)+1;
	  if(val<tmp[i*image->cols+j]){
	    tmp[i*image->cols+j]=val;
	  }
	}else{
	  val=downRightMin4(tmp,i,j,image->rows,image->cols)+1;
	  if(val<tmp[i*image->cols+j]){
	    tmp[i*image->cols+j]=val;
	  }
	}
      }
    }
  }



  for(i=0;i<image->rows;i++){
    for(j=0;j<image->cols;j++){
      if(tmp[i*image->cols+j]<thresh){
	if(!invert){
	  image->image[i*image->cols+j]=bgVal;
	}else{
	  image->image[i*image->cols+j]=(Pixel){255,255,255};
	}
      }
    }
  }
  free(tmp);
}

